from typing import Iterable

from yandil.discovery.class_discovery import (
    ClassData,
    discover_classes_from_module,
    exclude_classes_without_decorators,
    transform_class_nodes_to_class_data,
)
from yandil.discovery.module_discovery import discover_modules


class DeclarativeDependencyLoader:
    def __init__(
        self,
        discovery_base_path: str,
        sources_root_path: str,
    ):
        self.__sources_root_path = sources_root_path
        self.__discovery_base_path = discovery_base_path

    def load(self) -> None:
        for module_data in discover_modules(self.__discovery_base_path, set()):
            module_file_path = module_data.module_path + ".py"
            for class_data in self.__get_classes_discover_iterable(
                module_file_path,
            ):
                class_data.to_class(self.__sources_root_path)

    def __get_classes_discover_iterable(
        self,
        module_file_path: str,
    ) -> Iterable[ClassData]:
        iterable = exclude_classes_without_decorators(discover_classes_from_module(module_file_path), {"dependency"})
        return transform_class_nodes_to_class_data(iterable, module_file_path)



from typing import Any, Dict, Optional, Type, get_type_hints

from yandil.container import Container, default_container


class DependencyFiller:
    def __init__(self, container: Optional[Container] = None):
        if container is None:
            container = default_container
        self.__dependencies_container = container

    def fill(self, target_class: Type) -> None:
        target_class_base_init = target_class.__init__
        dependencies = self.__get_target_class_dependencies(target_class)

        def init_with_dependencies(target_self) -> None:
            target_class_base_init(target_self, **dependencies)

        target_class.__init__ = init_with_dependencies

    def __get_target_class_dependencies(self, target_class: Type) -> Dict[str, Any]:
        dependencies_dict = {}
        for argument_name, argument_type in get_type_hints(target_class.__init__).items():
            dependencies_dict[argument_name] = self.__dependencies_container[argument_type]
        return dependencies_dict



from typing import Type


class DependencyNotFoundError(Exception):
    def __init__(self, dependency_type: Type):
        self.dependency_type = dependency_type
        super().__init__(f"Dependency {dependency_type.__name__} not found")



from functools import wraps
from typing import Optional, TypeVar

from yandil.container import Container, default_container

DT = TypeVar("DT")


def dependency(
    cls: Optional[DT] = None, /, *, container: Optional[Container] = None, is_primary: Optional[bool] = None
):
    if container is None:
        container = default_container

    @wraps(cls)
    def wrap(cls: DT) -> DT:
        container.add(cls, is_primary=is_primary)
        return cls

    if cls is None:
        return wrap

    return wrap(cls)



class MissingConfigurationValueError(Exception):
    def __init__(self, configuration_value_name: str):
        self.configuration_value_name = configuration_value_name
        super().__init__(f"Missing configuration value for {configuration_value_name}")



from pathlib import Path
from typing import Iterable, Optional, Set

from yandil.container import Container, default_container
from yandil.discovery.class_discovery import (
    ClassData,
    discover_classes_from_module,
    exclude_abstract_classes,
    exclude_classes_without_public_methods,
    exclude_dataclasses,
    transform_class_nodes_to_class_data,
)
from yandil.discovery.module_discovery import discover_modules


class SelfDiscoverDependencyLoader:
    def __init__(
        self,
        discovery_base_path: str,
        sources_root_path: str,
        container: Optional[Container] = None,
        excluded_modules: Optional[Set[str]] = None,
        should_exclude_classes_without_public_methods: bool = True,
        should_exclude_dataclasses: bool = True,
        mandatory_modules: Optional[Set[Path]] = None,
    ):
        if container is None:
            container = default_container
        self.__container = container
        self.__excluded_modules = excluded_modules
        self.__should_exclude_classes_without_public_methods = should_exclude_classes_without_public_methods
        self.__should_exclude_dataclasses = should_exclude_dataclasses
        self.__sources_root_path = sources_root_path
        self.__discovery_base_path = discovery_base_path
        if mandatory_modules is None:
            mandatory_modules = set()
        self.__mandatory_modules = mandatory_modules

    def load(self) -> None:
        for module_data in discover_modules(self.__discovery_base_path, self.__excluded_modules):
            module_file_path = module_data.module_path + ".py"

            if self.__is_module_mandatory(module_data.module_path):
                module_classes = self.__get_mandatory_module_classes_discover_iterable(module_file_path)
            else:
                module_classes = self.__get_non_mandatory_module_classes_discover_iterable(
                    module_file_path,
                    self.__should_exclude_classes_without_public_methods,
                    self.__should_exclude_dataclasses,
                )

            for class_data in module_classes:
                cls = class_data.to_class(self.__sources_root_path)
                self.__container.add(cls)

    def __is_module_mandatory(self, module_path: str) -> bool:
        module_path = Path(module_path)
        for mandatory_module_path in self.__mandatory_modules:
            if module_path == mandatory_module_path:
                return True
            if mandatory_module_path in module_path.parents:
                return True
        return False

    def __get_mandatory_module_classes_discover_iterable(self, module_file_path: str) -> Iterable[ClassData]:
        return transform_class_nodes_to_class_data(
            class_nodes=exclude_abstract_classes(discover_classes_from_module(module_file_path)),
            module_file_path=module_file_path,
        )

    def __get_non_mandatory_module_classes_discover_iterable(
        self,
        module_file_path: str,
        should_exclude_classes_without_public_methods: bool,
        should_exclude_dataclasses: bool,
    ) -> Iterable[ClassData]:
        iterable = exclude_abstract_classes(discover_classes_from_module(module_file_path))
        if should_exclude_classes_without_public_methods:
            iterable = exclude_classes_without_public_methods(iterable)
        if should_exclude_dataclasses:
            iterable = exclude_dataclasses(iterable)
        return transform_class_nodes_to_class_data(iterable, module_file_path)



class ConfigurationValueTypeMismatchError(Exception):
    def __init__(self, configuration_value_name: str, expected_type: type, actual_type: type):
        self.configuration_value_name = configuration_value_name
        self.expected_type = expected_type
        self.actual_type = actual_type
        super().__init__(
            f"Expected type for configuration value {configuration_value_name} "
            f"was {expected_type}, got {actual_type}"
        )
